{******************************************************************************}
{                                  TRERuler
{                                  --------
{
{                     Copyright  2004-2012 Pieter Zijlstra
{
{ TRERuler is a descendant of TRuler specialized for the standard TRichEdit.
{
{ E-mail: p.zylstra@hccnet.nl
{ Website: http://home.hccnet.nl/p.zylstra/
{===============================================================================
{                           This software is FREEWARE
{                           -------------------------
{
{ You may freely use it in any software, including commercial software, but
{ be-aware that the code is provided as-is, with no implied warranty.
{===============================================================================
{
{
{Notes:
{ The following table shows which dynamic-link library (DLL) corresponds with
{ which version of Rich Edit. Note that the name of the file did not change
{ from version 2.0 to version 3.0. This allows version 2.0 to be upgraded to
{ version 3.0 without breaking existing code.
{
{ Rich Edit version DLL
{ 1.0 Riched32.dll
{ 2.0 Riched20.dll
{ 3.0 Riched20.dll
{ 4.1 Msftedit.dll
{
{ The following list describes which versions of Rich Edit are included in
{ which releases of Microsoft Windows.
{ Windows XP SP1 Includes v4.1, v3.0 and a v1.0 emulator.
{ Windows XP     Includes v3.0 with a v1.0 emulator.
{ Windows Me     Includes v1.0 and v3.0.
{ Windows 2000   Includes v3.0 with a v1.0 emulator.
{ Windows NT 4.0 Includes v1.0 and v2.0.
{ Windows 98     Includes v1.0 and v2.0.
{ Windows 95     Includes v1.0. However, Riched20.dll is compatible with
{                               Windows 95 and may be installed by an
{                               application that requires it.
{===============================================================================
{
{ Version history:
{ 1.0.0.0 22 feb 2004 - First release.
{ 1.6.1.1 21 may 2004 - Changed the Parents of the fake margins to the same
{                       parent of the RichEdit (it used to be the TForm).
{ 1.6.1.2 03 aug 2004 - Added v3 enhanced tabs.
{ 1.7.6.0 13 aug 2007 - New: Added DragItem parameter to DoRulerItemMove(),
{                            DoRulerItemRelease() and DoRulerItemSelect() to
{                            give descendants more control on what to do when
{                            a ruler item is being moved.
{ 1.7.6.1 10 mar 2008 - Removed test code (enhanced tabs, it was buggy).
{                     - Fix: Changed TRichEdit to TCustomRichEdit.
{ 1.7.7.0 24 sep 2008 - Checked for D2009.
{ 1.7.8.0  1 sep 2009 - Checked for D2010.
{ 1.8.0.0 15 jan 2012 - Checked for DXE2.
{******************************************************************************}
unit RERuler;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms,
  ComCtrls, ExtCtrls, RichEdit, Ruler;

type
  TRERuler = class(TRuler)
  private
    FControlCanvas: TControlCanvas;
    FLineStyle: TPenStyle;
    FOrgSelectionChange: TNotifyEvent;
    FPanelLM: TPanel;
    FPanelRM: TPanel;
    FRichEdit: TCustomRichEdit;
    FSaveX: Integer;
    FUpdating: Boolean;
    procedure AssignRichEditEvents;
    procedure DrawMarkLine;
    procedure RestoreRichEditEvents;
    procedure RichEditSelectionChange(Sender: TObject);
    procedure SetRichEdit(const Value: TCustomRichEdit);
  protected
    procedure DoRulerItemMove(DragItem: TDragItem; X: Integer; Removing: Boolean); override;
    procedure DoRulerItemRelease(DragItem: TDragItem); override;
    procedure DoRulerItemSelect(DragItem: TDragItem; X: Integer); override;
    procedure DoIndentChanged; override;
    procedure DoMarginChanged; override;
    procedure DoTabChanged; override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetEditRect; virtual;
    procedure UpdateRulerIndents;
    procedure UpdateRulerMargins;
  published
    property LineStyle: TPenStyle read FLineStyle write FLineStyle default psDot;
    property RichEdit: TCustomRichEdit read FRichEdit write SetRichEdit;
  end;

implementation

{ TRERuler }

procedure TRERuler.AssignRichEditEvents;
begin
  if not Assigned(FRichEdit) or (csDesigning in ComponentState) then
    Exit;
  FOrgSelectionChange := TRichEdit(FRichEdit).OnSelectionChange;
  TRichEdit(FRichEdit).OnSelectionChange := RichEditSelectionChange;
end;

constructor TRERuler.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLineStyle := psDot;
end;

procedure TRERuler.DoIndentChanged;
var
  M: Extended;
begin
  inherited;
  if not Assigned(RichEdit) or FUpdating then
    Exit;
  M := MultiPoints;
  RichEdit.Paragraph.FirstIndent := Round(M * FirstIndent);
  RichEdit.Paragraph.LeftIndent  := Round(M * LeftIndent);
  RichEdit.Paragraph.RightIndent := Round(M * RightIndent);
  UpdateRulerIndents;
end;

procedure TRERuler.DoMarginChanged;
begin
  inherited;
  SetEditRect;
  UpdateRulerMargins;
end;

procedure TRERuler.DoRulerItemMove(DragItem: TDragItem; X: Integer; Removing: Boolean);
begin
  inherited;
  if not Assigned(RichEdit) then
    Exit;
  DrawMarkLine;
  if Removing then
    FSaveX := -1
  else
    FSaveX := X - Inset;
  DrawMarkLine;
end;

procedure TRERuler.DoRulerItemRelease(DragItem: TDragItem);
begin
  inherited;
  if not Assigned(RichEdit) then
    Exit;
  DrawMarkLine;
  FControlCanvas.Free;
  FControlCanvas := nil;
  FRichEdit.Refresh;
end;

procedure TRERuler.DoRulerItemSelect(DragItem: TDragItem; X: Integer);
begin
  inherited;
  if not Assigned(RichEdit) then
    Exit;
  RichEdit.Update;
  if not Assigned(FControlCanvas) then
    FControlCanvas := TControlCanvas.Create;
  with FControlCanvas do
  begin
    Control := RichEdit;
    Pen.Color := clBlack;
    Pen.Mode := pmXor;
    Pen.Style := LineStyle;
  end;
  FSaveX := X - Inset;
  DrawMarkLine;
end;

procedure TRERuler.DoTabChanged;
var
  I: Integer;
begin
  inherited;
  if not Assigned(RichEdit) or FUpdating then
    Exit;
  RichEdit.Paragraph.TabCount := Tabs.Count + 1; // Workaroud for TRichEdit Tabs bug.
  for I := 0 to Tabs.Count - 1 do
    RichEdit.Paragraph.Tab[I] := Round(Tabs[I].Points);
  RichEdit.Paragraph.TabCount := Tabs.Count;
end;

procedure TRERuler.DrawMarkLine;
begin
  if Assigned(FControlCanvas) then
    with FControlCanvas do
      if FSaveX >= 0 then begin
        MoveTo(FSaveX{-FRichEdit.HScrollPos}, 0);
        LineTo(FSaveX{-FRichEdit.HScrollPos}, RichEdit.Height);
      end;
end;

procedure TRERuler.Loaded;
begin
  inherited;
  AssignRichEditEvents;
end;

procedure TRERuler.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FRichEdit) then
  begin
    RestoreRichEditEvents;
    FRichEdit := nil;
  end;
end;

procedure TRERuler.RestoreRichEditEvents;
begin
  if not Assigned(FRichEdit) or (csDesigning in ComponentState) then
    Exit;
  TRichEdit(FRichEdit).OnSelectionChange := FOrgSelectionChange;
  FOrgSelectionChange := nil;
end;

procedure TRERuler.RichEditSelectionChange(Sender: TObject);
begin
  if Assigned(FOrgSelectionChange) then
    FOrgSelectionChange(Sender);
  UpdateRulerIndents;
end;

procedure TRERuler.SetEditRect;
var
  R: TRect;
begin
  if not Assigned(FRichEdit) or (csDesigning in ComponentState) then
    Exit;
  R := Rect(Round(UnitsToPixs(LeftMargin)), 0,
            Round(UnitsToPixs(PageWidth - RightMargin)), RichEdit.ClientHeight);
  // Correction for W2K and XP, dont' know if this needed for NT.
  if Win32Platform = VER_PLATFORM_WIN32_NT then
    InflateRect(R, -1, 0);
  SendMessage(RichEdit.Handle, EM_SETRECT, 0, LongInt(@R));
  if Win32Platform = VER_PLATFORM_WIN32_NT then
    InflateRect(R, +1, 0);

  // Draw some fake margin lines by using small panels.
  FPanelLM.Left := RichEdit.Left + 1 + R.Left;
  FPanelRM.Left := RichEdit.Left + 3 + R.Right;
end;

procedure TRERuler.SetRichEdit(const Value: TCustomRichEdit);

  procedure SetPanelDefaults(Panel: TPanel);
  begin
    Panel.Parent := RichEdit.Parent;
    Panel.BevelInner := bvNone;
    Panel.BevelOuter := bvNone;
    Panel.Top := RichEdit.Top + 2;
    Panel.Height := RichEdit.Height - 4;
    Panel.Width := 1;
  end;

begin
  RestoreRichEditEvents;
  FRichEdit := Value;
  if not (csLoading in ComponentState) then
    AssignRichEditEvents;

  if not Assigned(FRichEdit) or (csDesigning in ComponentState) then
    Exit;

  FPanelLM := TPanel.Create(FRichEdit);
  SetPanelDefaults(FPanelLM);

  FPanelRM := TPanel.Create(FRichEdit);
  SetPanelDefaults(FPanelRM);
end;


procedure TRERuler.UpdateRulerIndents;
var
  I: Integer;
  M: Extended;
begin
  if not Assigned(FRichEdit) or (csDesigning in ComponentState) then
    Exit;
  FUpdating := True;
  try
    M := 1 / MultiPoints;
    FirstIndent := M * RichEdit.Paragraph.FirstIndent;
    LeftIndent  := M * RichEdit.Paragraph.LeftIndent;
    RightIndent := M * RichEdit.Paragraph.RightIndent;

    // Update tabs
    Tabs.Clear;
    for I := 0 to RichEdit.Paragraph.TabCount - 1 do
      Tabs.Add.Points := RichEdit.Paragraph.Tab[I];
  finally
    FUpdating := False;
  end;
end;

procedure TRERuler.UpdateRulerMargins;
begin
//
end;

end.
